from pathlib import Path
from typing import List, Union, Type, Tuple

from spacy.language import Language
from spacy.syntax.nn_parser import Parser
from spacy.tokens.doc import Doc


class BaseDocReader(object):
    """
    A base class for document reader, define interfaces for subclasses to inherent from
    """

    def __init__(self, nlp: Language = None, **kwargs):
        """

        :param nlp: a SpaCy language model
        :param kwargs: other parameters
        """
        for param_name, value in kwargs.items():
            setattr(self, param_name, value)
        if nlp is None:
            raise NameError('parameter "nlp" need to be defined')
        self.nlp = nlp
        self.txt = None
        self.anno = None
        pass

    def get_txt_content(self, txt_file: Path) -> str:
        """
        May need to be overwritten, if the txt_file is not a txt file, e.g. an xmi file
        :param txt_file: the path of a text file
        :return: the content of text file
        """
        return Path(txt_file).read_text(encoding='UTF8')

    def get_anno_content(self, txt_file: Path) -> str:
        """
        Either this method or infer_anno_file_path must be overwritten, as different annotation format
        use different conventions to store annotation files and text files (different in relative paths, or use a
        single file to store both of them)
        :param txt_file: the path of a text file
        :return: the content of text file
        """
        anno_file = self.infer_anno_file_path(txt_file)
        self.check_file_validity(anno_file)
        return Path(anno_file).read_text(encoding='UTF8')

    def infer_anno_file_path(self, txt_file: Path) -> Path:
        """
        From the path of the text file, infer the corresponding annotation file. Need to be implemented for each subclass,
        as different annotation format use different conventions to store annotation files and txt files
        """
        return None

    def check_file_validity(self, file: Union[str, Path], raise_error: bool = True) -> Path:
        if file is None or (isinstance(file, str) and len(file) == 0):
            if raise_error:
                raise NameError(file)
            else:
                return None
        if isinstance(file, str):
            file = Path(file)
        if not file.exists() or not file.is_file():
            if raise_error:
                raise FileNotFoundError(file.absolute())
            else:
                return None
        return file

    def read(self, txt_file: Union[str, Path]) -> Doc:
        """
        Read annotations and return a Spacy Doc, need to be implemented in subclasses

        :param txt_file: the text file path in an annotation project, where the corresponding
         annotation file path can be inferred through method: infer_anno_file_path
        """
        self.get_contents(txt_file=txt_file)
        doc = self.nlp(self.txt)
        return self.process(doc)

    def process(self, doc):
        """:arg a SpaCy Doc, must be implemented in the subclass."""
        return doc

    def get_contents(self, txt_file: Union[str, Path]):
        txt_file = self.check_file_validity(txt_file)
        if txt_file is not None:
            self.txt = self.get_txt_content(txt_file)
            self.anno = self.get_anno_content(txt_file)
        pass

    def find_start_token(self, start: int, left_token_offset: int,
                         right_token_offset: int, doc: Doc) -> int:
        """
        Use binary search to find the token-based offset of an annotation (Span) start

        :arg start: the start character offset of input span
        :arg left_token_offset: the left boundary (token offset) of the search window
        :arg right_token_offset: the right boundary (token offset) of the search window
        :arg doc: the input SpaCy Doc
        """
        if right_token_offset >= left_token_offset:
            mid = int(left_token_offset + (right_token_offset - left_token_offset) / 2)
            if doc[mid].idx <= start and start < doc[mid].idx + len(doc[mid]):
                return mid
            elif doc[mid].idx > start:
                return self.find_start_token(start, left_token_offset, mid - 1, doc)
            else:
                return self.find_start_token(start, mid + 1, right_token_offset, doc)
        else:
            return -1

    def find_end_token(self, end: int, token_start: int, total: int, doc: Doc):
        """
        Assume most of the annotations are short (a few tokens), it will more efficient to not use binary search here.
        :arg end: the end character offset of input span
        :arg token_start: the start token offset of the input span
        :arg total: the total number of tokens--check if span'end are invalid
        :arg doc: SpaCy Doc
        """
        for i in range(token_start, total):
            if end <= doc[i].idx + len(doc[i]):
                return i + 1
        return -1


class BaseDirReader:
    """
    A base class for directory reader, define interfaces for subclasses to inherent from
    """

    def __init__(self, txt_dir: Union[str, Path], txt_extension: str = 'txt',
                 nlp: Language = None,
                 docReaderClass: Type = None, recursive: bool = False, **kwargs):
        """

        :param txt_dir: the directory contains text files (can be annotation file, if the text content and annotation content are saved in the same file).
        :param txt_extension: the text file extension name (default is 'txt').
        :param nlp: a SpaCy language model.
        :param docReaderClass: a DocReader class that can be initiated.
        :param recursive: whether read file recursively down to the subdirectories.
        :param kwargs: other parameters that need to pass on to this DirReader or docReaderClass above
        """
        self.txt_dir = self.check_dir_validity(txt_dir)
        self.txt_extension = txt_extension
        self.recursive = recursive
        if docReaderClass is None or not issubclass(docReaderClass, BaseDocReader):
            raise NameError('docReaderClass must not be None, and must be a subclass of "BaseDocReader."')
        if nlp is None:
            raise NameError('parameter "nlp" need to be defined')
        self.nlp = nlp
        self.kwargs = kwargs
        self.reader = docReaderClass(nlp=self.nlp, **self.kwargs)
        pass

    def check_dir_validity(self, dir: Union[str, Path], raise_error: bool = True) -> Path:
        if dir is None or (isinstance(dir, str) and len(dir) == 0):
            if raise_error:
                raise NameError(dir)
            else:
                return None
        if isinstance(dir, str):
            dir = Path(dir)
        if not dir.exists() or not dir.is_dir():
            if raise_error:
                raise NotADirectoryError(dir)
            else:
                return None
        return dir

    def read(self) -> List[Doc]:
        """
        Read text files and annotation files, return a list of SpaCy Doc, need to be implemented in subclasses
        """
        txt_files = self.list_files(self.txt_dir, self.txt_extension)
        docs = []
        for txt_file in txt_files:
            try:
                doc = self.reader.read(txt_file)
                docs.append(doc)
            except:
                raise IOError('An error occured while parsing annotation for document: {}'.format(txt_file.absolute()))
            pass
        return docs

    def list_files(self, input_dir_path: Path, file_extension: str) -> List[Path]:
        if self.recursive:
            files = list(input_dir_path.rglob('*.' + file_extension))
        else:
            files = list(input_dir_path.glob('*.' + file_extension))
        return files



import spacy
from spacy.tokens import Doc, Span, Token


class DocConsumer(object):
	"""A DocConsumer object will consume a spacy doc and output rows based on a configuration provided by the user."""
	def __init__(self, span_type='ent', labels=None, attrs=None, id_attr=None):
		""" Create a DocConsumer object

		Args:
			span_type: the type item to output. Can be {"doc", "ent", "token"}. Default is "ent".
			labels: if outputting entities, output entities with the specified .label_ properties.
				Default None, will output all entites.
			attrs: output the following attributes belonging to span_type. Default None, will output all attrs.
			id_attr: if the Doc object has an identifier that will be output with every row. Default None.
		"""
		super(DocConsumer, self).__init__()
		self.type_dict = {
			"doc":self.get_rows_doc,
			"ent":self.get_rows_ent,
			"token":self.get_rows_token,
			"rels":self.get_rows_rels # for outputting relations at future date
		}

		if span_type in ["doc", "ent", "token"]:
			self.span_type = span_type
		else:
			raise ValueError("Must provide a valid span_type: 'doc', 'ent' or 'token' not {0}".format(span_type))

		if labels:
			self.labels = set(labels)
		else:
			self.labels = None

		if attrs is None: # if no attributes specified, output all
			if span_type is 'ent':
				self.attrs = {"text", "label_", "start", "end"}
			elif span_type is 'token':
				self.attrs = {"text", "like_num", "is_punct"}
			else:
				self.attrs = {"text"}
		else: # otherwise output the specified attributes
			self.attrs = attrs

	def get_rows(self, doc):
		"""Get the rows of the specified document based on object's configs."""
		return self.type_dict[self.span_type](doc)

	def get_rows_doc(self, doc):
		row = []
		for attr in self.attrs:
			val = None
			try:
				val = getattr(doc,attr)
			except AttributeError:
				val = doc._.get(attr)
				#TODO: what happens when there's an exception here?
			row.append(val)
		return row


	def get_rows_ent(self, doc):
		rows = []
		for ent in doc.ents:
			row = []
			if not self.labels or ent.label_ in self.labels:
				for attr in self.attrs:
					val = None
					try:
						val = getattr(ent,attr)
					except AttributeError:
						val = ent._.get(attr)
						#TODO: what happens when there's an exception here?
					row.append(val)
				rows.append(row)
		return rows

	def get_rows_token(self, doc):
		rows = []
		for token in doc:
			row = []
			for attr in self.attrs:
				val = None
				try:
					val = getattr(token,attr)
				except AttributeError:
					val = token._.get(attr)
					#TODO: what happens when there's an exception here?
				row.append(val)
			rows.append(row)
		return rows

	def get_rows_rels(self, doc):
		raise NotImplementedError

def print_rows(rows):
	for row in rows:
		print(row)





# This package contains a list of annotation readers, which take in annotation files (with or without document text files) and construct a spacy doc
from .base_reader import BaseDocReader, BaseDirReader
from .ehost_reader import EhostDocReader, EhostDirReader



import spacy

from medspacy_io.consumer.doc_consumer import DocConsumer, print_rows

nlp = spacy.load("en_core_web_sm")



d_consumer = DocConsumer("doc")
ent_consumer = DocConsumer("ent")
token_consumer = DocConsumer("token")

# first few paragraphs from ny times article clicked from front page
# https://www.nytimes.com/2020/04/18/us/texas-protests-stay-at-home.html
doc = nlp("In a public act of defiance, several dozen protesters in Texas converged on the steps of the Capitol building in Austin on Saturday to call for the reopening of the state and the country during the coronavirus pandemic. The 'You Can’t Close America' rally rode a wave of similar protests at statehouses and in city streets this past week, with people also gathering on Saturday in Indianapolis; Carson City, Nev.; Annapolis, Md.; and Brookfield, Wis. With more than 22 million unemployment claims nationwide in the past four weeks, some conservatives have begun voicing displeasure with the moribund economy. Many businesses have been shuttered in an attempt to slow the spread of the virus, which has killed more than 34,000 people in the United States. The protests have been encouraged by President Trump, but polls show that most Americans support restrictions meant to combat the virus. This month, a poll by Quinnipiac University found that 81 percent of registered voters supported a theoretical nationwide stay-at-home order, including 68 percent of Republicans polled. Others, though, are openly violating the stay-at-home orders replicated by governors across the country by assembling to express their dissatisfaction. In Austin, at least 100 people gathered on the statehouse grounds in hats and shirts with President Trump’s slogan, “Make America Great Again.” Some carried American flags, and few wore masks that are mandated by the city. There were cheers at the sight of Alex Jones, the founder of the website Infowars, which traffics in conspiracy theories. There were chants of 'Fire Fauci,' in reference to Dr. Anthony S. Fauci, the federal government’s top infectious disease expert.")

print("CONSUMING DOCUMENT...")
print(list(d_consumer.attrs))
print_rows(d_consumer.get_rows(doc))

print("\nCONSUMING ENTITIES...")
print(list(ent_consumer.attrs))
print_rows(ent_consumer.get_rows(doc))

print("\nCONSUMING TOKENS...")
print(list(token_consumer.attrs))
print_rows(token_consumer.get_rows(doc))


from collections import OrderedDict
from pathlib import Path
from typing import Union, Tuple, Type

from lxml import etree
from lxml.etree import Element, iterparse
from spacy.language import Language
from spacy.tokens.span import Span

from reader import BaseDocReader, BaseDirReader


class EhostDocReader(BaseDocReader):
    """ This is a subclass of BaseDocReader to read in eHOST format files and generate SpaCy Docs
    """

    def __init__(self, nlp: Language = None, use_adjudication: bool = False,
                 schema_file: Union[str, Path] = ''):
        """

        :param nlp: a SpaCy language model
        :param use_adjudication: if read annotations from adjudication folder
        :param schema_file: initiate Span attributes using eHOST schema configuration file
        """
        self.schema_set = False
        self.set_attributes(schema_file=schema_file)

        super().__init__(nlp=nlp, use_adjudication=use_adjudication)
        pass

    def set_attributes(self, schema_file: Union[str, Path] = ''):
        """
        The current version SpaCy doesn't differentiate attributes for different annotation types.
        Thus, any attributes extended here will be applied to all Spans.
        :arg schema_file: initiate Span attributes using eHOST schema configuration file

        """
        schema_file = self.check_file_validity(schema_file, False)
        attr_names = set()
        if schema_file is not None:
            root = etree.parse(str(schema_file.absolute()))
            for attr_def in root.iter("attributeDef"):
                name = attr_def[0].text.replace(' ', '_')
                default_value = attr_def[2].text
            if name not in attr_names and not Span.has_extension(name):
                Span.set_extension(name, default=default_value)
                attr_names.add(name)
            self.schema_set = True
        pass

    def get_anno_content(self, txt_file: Path) -> str:
        """
        :arg txt_file from the text file path, infer knowtator.xml file path
        To use lxml's iterparse, just keep the path string here instead of reading the content
        """
        anno_file = self.infer_anno_file_path(txt_file)
        self.check_file_validity(anno_file)
        return str(anno_file.absolute())

    def infer_anno_file_path(self, txt_file: Path) -> Path:
        txt_file_name = txt_file.name
        anno_file_name = txt_file_name + '.knowtator.xml'
        if not self.use_adjudication:
            anno_file = Path(txt_file.parent.parent, 'saved', anno_file_name)
        else:
            anno_file = Path(txt_file.parent.parent, 'adjudication', anno_file_name)
        self.check_file_validity(anno_file)
        return anno_file

    def process(self, doc):
        """
        Take in a SpaCy doc, add eHOST annotation(Span)s.
         :arg doc: a SpaCy doc
        """
        spans, classes, attributes = self.parse_to_dicts(self.anno)
        sorted_span = sorted(spans.items(), key=lambda x: x[1][0])

        existing_entities = list(doc.ents)
        new_entities = list()
        previous_token_offset = 0
        total = len(doc)
        for id, (start, end) in sorted_span:
            # because SpaCy uses token offset instead of char offset to define Spans, we need to match them,
            # binary search is used here to speed up
            token_start = self.find_start_token(start, previous_token_offset, total, doc)
            token_end = self.find_end_token(end, token_start, total, doc)
            if token_start >= 0 and token_end > 0:
                span = Span(doc, token_start, token_end, label=classes[id][0])
                for attr_id in classes[id][1]:
                    attr_name = attributes[attr_id][0]
                    attr_value = attributes[attr_id][1]
                    setattr(span._, attr_name, attr_value)
                new_entities.append(span)
                previous_token_offset = token_end
            else:
                raise OverflowError(
                    'The span of the annotation: {}[{}:{}] is out of document boundary.'.format(classes[id][0], start,
                                                                                                end))
            pass
        doc.ents = existing_entities + new_entities
        return doc

    def parse_to_dicts(self, xml_file: str) -> Tuple[OrderedDict, OrderedDict]:
        iter = etree.iterparse(xml_file, events=('start',))
        # this doesn't seem elegant, but is said to be the fastest way
        spans = OrderedDict()
        classes = dict()
        attributes = OrderedDict()
        for event, ele in iter:
            if ele.tag == 'annotation':
                id, start, end, span_text = self.parse_annotation_tag(ele, iter)
                spans[id] = (start, end)
            elif ele.tag == 'stringSlotMention':
                attr_id, attr, value = self.parse_attribute_tag(ele, iter)
                attributes[attr_id] = (attr, value)
            elif ele.tag == 'classMention':
                # <classMention id="EHOST_Instance_1">
                #   <hasSlotMention id="EHOST_Instance_8"/>
                #   <mentionClass id="Purulent">pain</mentionClass>
                # some annotations don't have "hasSlotMention" element
                # </classMention>
                id = ele.get('id')
            elif ele.tag == 'mentionClass':
                class_tag = ele.get('id')
                if id not in classes:
                    classes[id] = ['', []]
                classes[id][0] = class_tag
            elif ele.tag == 'hasSlotMention':
                if id not in classes:
                    classes[id] = ['', []]
                classes[id][1].append(ele.get('id'))
        return spans, classes, attributes

    # <annotation>
    #   <mention id="EHOST_Instance_1"/>
    #   <annotator id="eHOST_2010">sjl</annotator>
    #   <span start="232" end="236"/>
    #   <spannedText>pain</spannedText>
    #   <creationDate>Sun Apr 19 19:30:11 MDT 2020</creationDate>
    # </annotation>
    def parse_annotation_tag(self, ele: Element, iter: iterparse) -> Tuple:
        id = None
        start = -1
        end = -1
        span_text = None
        for i in range(0, 4):
            eve, child = iter.__next__()
            if child.tag == 'mention':
                id = child.get('id')
            elif child.tag == 'span':
                start = int(child.get('start'))
                end = int(child.get('end'))
            elif child.tag == 'spannedText':
                span_text = child.text
        return id, start, end, span_text

    # <stringSlotMention id="EHOST_Instance_8">
    #   <mentionSlot id="status"/>
    #   <stringSlotMentionValue value="present"/>
    # </stringSlotMention>
    def parse_attribute_tag(self, ele: Element, iter: iterparse) -> Tuple:
        id = ele.get('id')
        attr = ''
        value = ''
        for i in range(0, 2):
            eve, child = iter.__next__()
            if child.tag == 'mentionSlot':
                attr = child.get('id')
            else:
                value = child.get('value')
        return id, attr, value


class EhostDirReader(BaseDirReader):
    def __init__(self, txt_dir: Union[str, Path], txt_extension: str = 'txt',
                 nlp: Language = None, docReaderClass: Type = None,
                 recursive: bool = False, use_adjudication: bool = False,
                 schema_file: Union[str, Path] = ''):
        """
        :param txt_dir: the directory contains text files (can be annotation file, if the text content and annotation content are saved in the same file).
        :param txt_extension: the text file extension name (default is 'txt').
        :param nlp: a SpaCy language model
        :param docReaderClass: a DocReader class that can be initiated.
        :param recursive: whether read file recursively down to the subdirectories.
        :param use_adjudication: if read annotations from adjudication folder
        :param schema_file: initiate Span attributes using eHOST schema configuration file
        """
        super().__init__(txt_dir=txt_dir, txt_extension=txt_extension, nlp=nlp, docReaderClass=docReaderClass,
                         recursive=recursive, use_adjudication=use_adjudication, schema_file=schema_file)
        pass



from setuptools import setup
from codecs import open
from os import path
from setuptools.extension import Extension

here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


def get_version():
    """Load the version from version.py, without importing it.

    This function assumes that the last line in the file contains a variable defining the
    version string with single quotes.

    """
    try:
        with open('medspacy_io/version.py', 'r') as f:
            return f.read().split('\n')[0].split('=')[-1].replace('\'', '').strip()
    except IOError:
        return "0.0.0a1"


dir_path = path.dirname(path.realpath(__file__))
include_dirs = [dir_path + "/medspacy", dir_path]

setup(
    name='medspacy_io',
    packages=['medspacy_io', 'medspacy_io.reader', 'medspacy_io.writer', 'medspacy_io.consumer'],
    # this must be the same as the name above
    version=get_version(),
    description='A collection of modules to facilitate reading text from various sources and writing to various sources.',
    author="medSpaCy",
    author_email="medspacy.dev@gmail.com",
    url='https://github.com/medspacy/medspacy_io',  # use the URL to the github repo
    keywords=['medspacy', 'reader', 'writer', 'ehost', 'brat', 'xmi', 'io', 'reader', 'writer', 'nlp', 'annotation'],
    long_description=long_description,
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Development Status :: 3 - Alpha',
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Text Processing :: Linguistic",
    ],
    license='MIT License',
    zip_safe=False,
    install_requires=[
        'lxml>=4.4.0', 'spacy>=2.2.2'
    ],
    setup_requires=[
        'lxml>=4.4.0', 'spacy>=2.2.2'
    ],
    test_suite='nose.collector',
    tests_require='nose',
    data_files=[('test_data', ['tests/data/ehost_test_corpus/config/projectschema.xml',
                               'tests/data/ehost_test_corpus/corpus/doc1.txt',
                               'tests/data/ehost_test_corpus/corpus/doc2.txt',
                               'tests/data/ehost_test_corpus/saved/doc1.txt.knowtator.xml',
                               'tests/data/ehost_test_corpus/saved/doc2.txt.knowtator.xml'])],
)



import os, shutil, sys
from lxml import etree
from lxml.etree import Element

from datetime import date

DEFAULT_COLOR = (220, 220, 220)

base_schema_string = """<?xml version="1.0" encoding="UTF-8"?>
<eHOST_Project_Configure Version="1.0">
    <Handling_Text_Database>false</Handling_Text_Database>
    <OracleFunction_Enabled>false</OracleFunction_Enabled>
    <AttributeEditor_PopUp_Enabled>false</AttributeEditor_PopUp_Enabled>
    <OracleFunction>true</OracleFunction>
    <AnnotationBuilder_Using_ExactSpan>false</AnnotationBuilder_Using_ExactSpan>
    <OracleFunction_Using_WholeWord>true</OracleFunction_Using_WholeWord>
    <GraphicAnnotationPath_Enabled>true</GraphicAnnotationPath_Enabled>
    <Diff_Indicator_Enabled>true</Diff_Indicator_Enabled>
    <Diff_Indicator_Check_CrossSpan>true</Diff_Indicator_Check_CrossSpan>
    <Diff_Indicator_Check_Overlaps>false</Diff_Indicator_Check_Overlaps>
    <StopWords_Enabled>false</StopWords_Enabled>
    <Output_VerifySuggestions>false</Output_VerifySuggestions>
    <Pre_Defined_Dictionary_DifferentWeight>false</Pre_Defined_Dictionary_DifferentWeight>
    <PreAnnotated_Dictionaries Owner="NLP_Assistant" />
    <attributeDefs />
    <Relationship_Rules />
    <classDefs />
</eHOST_Project_Configure>
"""

base_annotations_string = """<?xml version="1.0" encoding="UTF-8"?>
<annotations textSource="0_10814621490">
    <eHOST_Adjudication_Status version="1.0">
        <Adjudication_Selected_Annotators version="1.0" />
        <Adjudication_Selected_Classes version="1.0" />
        <Adjudication_Others>
            <CHECK_OVERLAPPED_SPANS>false</CHECK_OVERLAPPED_SPANS>
            <CHECK_ATTRIBUTES>false</CHECK_ATTRIBUTES>
            <CHECK_RELATIONSHIP>false</CHECK_RELATIONSHIP>
            <CHECK_CLASS>false</CHECK_CLASS>
            <CHECK_COMMENT>false</CHECK_COMMENT>
        </Adjudication_Others>
    </eHOST_Adjudication_Status>
</annotations>"""

class EhostWriter:

    def __init__(self, annotation_labels=None, annotation_color_mapping=None, context=True, sections=True,
                 span_attributes=None):
        if annotation_labels is None:
            annotation_labels = set()
        self.annotation_labels = annotation_labels
        self.color_cycle = self._create_color_generator()
        if annotation_color_mapping is None:
            annotation_color_mapping = self.create_default_color_mapping(annotation_labels)
        self.annotation_color_mapping = annotation_color_mapping

        self.context = context
        self.sections = sections
        if span_attributes is None:
            span_attributes = {
                'is_family',
                 'is_historical',
                 'is_hypothetical',
                 'is_negated',
                 'is_uncertain',
            }
        self.span_attributes = span_attributes

        self._i = 0

    def write_docs_to_ehost(self, docs, report_ids, directory, overwrite_directory=True):
        self.prepare_directory(directory, overwrite_directory)
        self.create_schema_file(directory)

        # Now save the texts
        for (doc, report_id) in zip(docs, report_ids):
            self.save_text(doc, report_id, directory)
            xml = self.create_ehost_xml(doc)
            self.save_ehost_xml(xml, report_id, directory)

    def prepare_directory(self, directory, overwrite_directory=True):
        # First, make the parent directory
        if not os.path.exists(directory):
            os.mkdir(directory)
        else:
            if overwrite_directory:
                shutil.rmtree(directory)
            os.mkdir(directory)

        # Now create the three subdirectories
        corpus_dir = os.path.join(directory, "corpus")
        xmls_dir = os.path.join(directory, "xmls")
        saved_dir = os.path.join(directory, "saved")
        config_dir = os.path.join(directory, "config")

        os.mkdir(corpus_dir)
        os.mkdir(xmls_dir)
        os.mkdir(saved_dir)
        os.mkdir(config_dir)

    def create_schema_file(self, root_dir):
        filepath = os.path.join(root_dir, "config", "projectschema.xml")
        xml = etree.fromstring(base_schema_string.encode())
        class_defs = xml.find('classDefs')
        for label, rgb in self.annotation_color_mapping.items():
            self.add_classdef(class_defs, label, rgb)
        with open(filepath, 'wb') as f:
            f.write(etree.tostring(xml, pretty_print=True))
        # print(f"Saved project config at {outpath}")

    def add_classdef(self, class_defs, label, color):
        class_def = Element('classDef')
        name = Element('Name', )
        name.text = label
        class_def.append(name)

        self.add_rgb_tags(class_def, color)

        # Add other tags
        sub = Element('InHerit_Public_Attributes')
        sub.text = 'true'
        class_def.append(sub)

        sub = Element('Source')
        sub.text = 'eHOST'
        class_def.append(sub)

        # Now add the attributes defined in span_attributes
        for attr in self.span_attributes:
            print(attr)
            attribute_def = Element("attributeDef")
            attribute_def.append(Element("Name", value=attr))
            attribute_def.append(Element("defaultValue", value="false"))
            attribute_def.append(Element("attributeDefOptionDef", value="false"))
            attribute_def.append(Element("attributeDefOptionDef", value="true"))
            class_def.append(attribute_def)

        class_defs.append(class_def)

    #     < attributeDef >
    #     < Name > is_negated < / Name >
    #     < is_Linked_to_UMLS_CUICode_and_CUILabel > false < / is_Linked_to_UMLS_CUICode_and_CUILabel >
    #     < is_Linked_to_UMLS_CUICode > false < / is_Linked_to_UMLS_CUICode >
    #     < is_Linked_to_UMLS_CUILabel > false < / is_Linked_to_UMLS_CUILabel >
    #     < defaultValue > false < / defaultValue >
    #     < attributeDefOptionDef > false < / attributeDefOptionDef >
    #     < attributeDefOptionDef > true < / attributeDefOptionDef >
    #
    # < / attributeDef >

    def add_rgb_tags(self, class_def, rgb):
        # Add RGB values
        for (value, name) in zip(rgb, ('RGB_R', 'RGB_G', 'RGB_B')):
            sub = Element(name)
            sub.text = str(value)
            class_def.append(sub)

    def save_text(self, doc, report_id, root_dir):
        corpus_dir = os.path.join(root_dir, "corpus")
        filepath = os.path.join(corpus_dir, "{}.txt".format(report_id))
        with open(filepath, "w") as f:
            f.write(doc.text)

    def create_ehost_xml(self, doc):
        xml = etree.fromstring(base_annotations_string.encode())
        # Start with the entity spans
        for ent in doc.ents:
            # Start with the span attributes
            slot_mention_ids = []

            for attr in self.span_attributes:
                self.add_string_slot_mention(ent, attr, xml, self._i)
                slot_mention_ids.append(self._i)
                self._i += 1


            if ent.label_ not in self.annotation_color_mapping:
                self.annotation_color_mapping[ent.label_] = DEFAULT_COLOR
            self.add_annotation(ent, xml, self._i)
            self.add_class_mention(ent, ent.label_, xml, self._i, slot_mention_ids)
            self._i += 1

        # Now add optional context modifiers and section headers
        if self.context is True and hasattr(doc._, "context_graph"):
            for _, modifier in doc._.context_graph.edges:
                if modifier.category not in self.annotation_color_mapping:
                    self.annotation_color_mapping[modifier.category] = DEFAULT_COLOR
                self.add_annotation(modifier.span, xml, self._i)
                self.add_class_mention(modifier.span, modifier.category, xml, self._i, [])
                self._i += 1
        if self.sections is True and hasattr(doc._, "section_headers"):
            for header in doc._.section_headers:
                if header is None:
                    continue
                if header._.section_title not in self.annotation_color_mapping:
                    self.annotation_color_mapping[header._.section_title] = DEFAULT_COLOR
                self.add_annotation(header, xml, self._i)
                self.add_class_mention(header, header._.section_title, xml, self._i, [])
                self._i += 1

        return xml

    def save_ehost_xml(self, xml, report_id, root_dir):
        saved_dir = os.path.join(root_dir, "saved")
        filepath = os.path.join(saved_dir, "{}.txt.knowtator.xml".format(report_id))
        with open(filepath, 'wb') as f:
            f.write(etree.tostring(xml, pretty_print=True))

    def add_annotation(self, span, xml, i):
        annotation = Element('annotation')
        # mention id
        mention = Element('mention', id=f'EHOST_Instance_{i}')
        annotation.append(mention)

        # annotator id
        annotator = Element('annotator', id='medSpaCy')
        annotator.text = 'medSpaCy'
        annotation.append(annotator)

        # span
        start = span.start_char
        end = start + len(span.text)
        span = Element('span', start=str(start), end=str(end))
        annotation.append(span)

        # spannedText
        spanned_text = Element('spannedText')
        spanned_text.text = span.text
        annotation.append(spanned_text)


        # creationDate
        # TODO

        xml.append(annotation)

    def add_class_mention(self, span, label, xml, i, slot_mention_ids):
        mention = Element('classMention', id=f'EHOST_Instance_{i}')

        mention_class = Element('mentionClass', id=label)
        mention_class.text = span.text
        mention.append(mention_class)

        xml.append(mention)

        # Add references to the slot mentions which represent various attributes
        for slot_mention_id in slot_mention_ids:
            slot_mention_ref = Element("hasSlotMention", id=f"EHOST_Instance_{self._i}")
            mention.append(slot_mention_ref)

    def add_string_slot_mention(self, span, attr, xml, i):
        value = str(getattr(span._, attr)).lower()

        slot_mention = Element("stringSlotMention", id=f'EHOST_Instance_{i}')
        mention_slot = Element("mentionSlot", id=attr)
        slot_mention.append(mention_slot)
        mention_slot_value = Element("mentionSlotValue", value=value)
        slot_mention.append(mention_slot_value)

        xml.append(slot_mention)

        #     < stringSlotMention
        #     id = "EHOST_Instance_327" >
        #     < mentionSlot
        #     id = "is_negated" / >
        #     < stringSlotMentionValue
        #     value = "true" / >
        #
        # < / stringSlotMention >

    def create_default_color_mapping(self, labels):
        color_mapping = dict()
        for label in labels:
            hex_color = next(self.color_cycle)
            rgb = self._hex_to_rgb(hex_color)
            color_mapping[label] = rgb
        return color_mapping

    def _create_color_generator(self):
        """Create a generator which will cycle through a list of default matplotlib colors"""
        from itertools import cycle
        colors = [u'#1f77b4', u'#ff7f0e', u'#2ca02c', u'#d62728',
                  u'#9467bd', u'#8c564b', u'#e377c2', u'#7f7f7f', u'#bcbd22', u'#17becf']
        return cycle(colors)

    def _hex_to_rgb(self, h):
        h = h.strip('#')
        return tuple(int(h[i:i + 2], 16) for i in (0, 2, 4))
